#
"""test_human_time - Test human readable time functions"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from datetime import datetime, timedelta, timezone

from pytest import mark, raises

from jnrbase.human_time import human_timestamp, parse_timedelta


human_timestamp_examples = [
    ({'days': 365, }, 'last year'),
    ({'days': 70, }, 'about two months ago'),
    ({'days': 30, }, 'last month'),
    ({'days': 21, }, 'about three weeks ago'),
    ({'days': 4, }, 'about four days ago'),
    ({'days': 1, }, 'yesterday'),
    ({'hours': 5, }, 'about five hours ago'),
    ({'hours': 1, }, 'about an hour ago'),
    ({'minutes': 6, }, 'about six minutes ago'),
    ({'seconds': 12, }, 'about 12 seconds ago'),
    ({}, 'right now'),
]


@mark.parametrize('delta,result', human_timestamp_examples)
def test_human_timestamp(delta, result):
    now = datetime.utcnow().replace(tzinfo=timezone.utc)
    dt = now - timedelta(**delta)
    assert human_timestamp(dt) == result


@mark.parametrize('delta,result', human_timestamp_examples)
def test_human_timestamp_naive(delta, result):
    dt = datetime.utcnow() - timedelta(**delta)
    assert human_timestamp(dt) == result


@mark.parametrize('string,dt', [
    ('3h', timedelta(0, 10800)),
    ('1d', timedelta(1)),
    ('1 d', timedelta(1)),
    ('0.5 y', timedelta(182, 43200)),
    ('0.5 Y', timedelta(182, 43200)),
])
def test_parse_timedelta(string, dt):
    assert parse_timedelta(string) == dt


def test_parse_invalid_timedelta():
    with raises(ValueError, match='Invalid ‘frequency’ value'):
        parse_timedelta('1 k')



#
"""test_colourise - Test colourisation functions"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from pytest import mark

from jnrbase import colourise


@mark.parametrize('f,expected', [
    (colourise.info, '\x1b[34m\x1b[1m'),
    (colourise.fail, '\x1b[31m\x1b[1m'),
    (colourise.success, '\x1b[32m\x1b[1m'),
    (colourise.warn, '\x1b[33m\x1b[1m'),
])
def test_colouriser(f, expected):
    assert expected in f('test')


def test_disabled_colouriser(monkeypatch):
    monkeypatch.setattr('jnrbase.colourise.COLOUR', False)
    assert colourise.info('test') == 'test'



#
"""i18n - Configure internationalisation support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

import gettext

from os import path
from types import ModuleType
from typing import Callable, Tuple


def setup(__pkg: ModuleType) -> Tuple[Callable[[str], str],
                                      Callable[[str, str, int], str]]:
    """Configure gettext for given package.

    Args:
        __pkg: Package to use as location for :program:`gettext` files
    Returns:
        :program:`gettext` functions for singular and plural translations

    """
    package_locale = path.join(path.dirname(__pkg.__file__), 'locale')
    gettext.install(__pkg.__name__, package_locale)

    return gettext.gettext, gettext.ngettext



#
"""test_pager - Test pager piping support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from shutil import which

from pytest import mark

from jnrbase.pager import pager


@mark.skipif(not which('cat'), reason='Requires cat')
def test_pager(monkeypatch, capfd):
    pager('paging through cat', pager='cat')
    assert capfd.readouterr()[0] == 'paging through cat'


def test_disable_pager(capsys):
    pager('pager forcibly disabled', pager=None)
    assert capsys.readouterr()[0] == 'pager forcibly disabled\n'



#
"""utils - Utility functions for tests."""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+


def func_attr(name, value):
    """Decorator to set an attribute on a function.

    Args:
        name (str): Attribute name to set
        value: Value to set attribute to
    """
    def decorator(f):
        setattr(f, name, value)
        return f
    return decorator



#
"""timer - Function timing support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from datetime import datetime

from .human_time import human_timestamp


class Timing:  # pylint: disable=too-few-public-methods

    """Timing context manager.

    .. versionchanged:: 0.3.0

        Renamed from ``Timer``.

    Attributes:
        verbose (bool): Print elapsed time
        elapsed (datetime.timedelta): Duration of execution
    """

    def __init__(self, *, verbose: bool = False) -> None:
        """Configure the timing Timing context manager.

        Args:
            verbose: Print elapsed time
        """
        self.verbose = verbose
        self._start = None
        self.elapsed = None

    def __enter__(self) -> 'Timing':
        self._start = datetime.utcnow()
        return self

    def __exit__(self, *args) -> None:
        now = datetime.utcnow()
        self.elapsed = now - self._start
        if self.verbose:
            print('Started {}'.format(human_timestamp(self._start)))



#
"""path_context - Environment modifying context handlers support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

import contextlib
import os

from typing import ContextManager, Dict, Union


@contextlib.contextmanager
def chdir(__path: str) -> ContextManager:
    """Context handler to temporarily switch directories.

    Args:
        __path: Directory to change to

    Yields:
        Execution context in ``path``
    """
    old = os.getcwd()
    try:
        os.chdir(__path)
        yield
    finally:
        os.chdir(old)


@contextlib.contextmanager
def env(**kwargs: Union[Dict[str, str], None]) -> ContextManager:
    """Context handler to temporarily alter environment.

    If you supply a value of ``None``, then the associated key will be deleted
    from the environment.

    Args:
        kwargs: Environment variables to override

    Yields:
        Execution context with modified environment
    """
    old = os.environ.copy()
    try:
        os.environ.clear()
        # This apparent duplication is because putenv doesn’t update
        # os.environ, and os.environ changes aren’t propagated to subprocesses.
        for key, value in old.items():
            os.environ[key] = value  # NOQA: B003
            os.putenv(key, value)
        for key, value in kwargs.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value  # NOQA: B003
                os.putenv(key, value)
        yield
    finally:
        os.environ.clear()
        for key, value in old.items():
            os.environ[key] = value  # NOQA: B003
            os.putenv(key, value)



#
"""conftest - Configuration for tests."""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from pytest import fixture


def _get_module(request):
    modbase = request.module.__name__.replace('tests.test_', '')
    return 'jnrbase.{}'.format(modbase)


@fixture
def path_exists_force(request, monkeypatch):
    """Force result of module's ``path.exists`` function.

    This fixture returns ``True`` by default, but a custom value can be
    specified by setting the ``exists_result`` attribute on a test function.
    If the ``exists_result`` attribute is a list it returns items in LIFO order
    on each call.
    """
    result = getattr(request.function, 'exists_result', True)
    if isinstance(result, list):
        monkeypatch.setattr('.'.join([_get_module(request), 'path', 'exists']),
                            lambda s: result.pop())
    else:
        monkeypatch.setattr('.'.join([_get_module(request), 'path', 'exists']),
                            lambda s: result)



#
"""test_attrdict - Test AttrDict support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from unittest import TestCase

from pytest import raises

from jnrbase.attrdict import AttrDict, ROAttrDict  # NOQA: F401


class AttrDictTest(TestCase):
    def setup_method(self, method):
        self.ad = globals()[self.__class__.__name__[:-4]](carrots=3, snacks=0)

    def test_base(self):
        assert isinstance(self.ad, dict)

        assert self.ad['carrots'] == 3
        assert self.ad['snacks'] == 0

        assert sorted(self.ad.keys()) == ['carrots', 'snacks']

    def test___contains__(self):
        assert 'carrots' in self.ad
        assert 'prizes' not in self.ad

    def test___getattr__(self):
        assert self.ad.carrots == 3
        assert self.ad.snacks == 0

    def test___setattr__(self):
        self.ad.carrots, self.ad.snacks = 0, 3
        assert self.ad.carrots == 0
        assert self.ad.snacks == 3

    def test___delattr__(self):
        assert 'carrots' in self.ad
        del self.ad['carrots']
        assert 'carrots' not in self.ad


class InvalidKeyTest(TestCase):
    def setup_method(self, method):
        self.ad = AttrDict(carrots=3, snacks=0)

    def test_invalid_key_set(self):
        with raises(AttributeError, match='unhashable'):
            self.ad.__setattr__({True: False}, None)

    def test_invalid_key_delete(self):
        with raises(AttributeError, match='unhashable'):
            self.ad.__delattr__({True: False})


class HasAttrTest(TestCase):
    def setup_method(self, method):
        self.ad = AttrDict(carrots=3, snacks=0)

        def raise_error():
            raise ValueError()
        self.ad.prop = property(raise_error)

    def test_swallowed_exception(self):
        assert 'prop' in self.ad


class ROAttrDictTest(AttrDictTest):
    def test___setattr__(self):
        with raises(AttributeError, match='is read-only'):
            self.ad.carrots = 1

    def test___delattr__(self):
        with raises(AttributeError, match='is read-only'):
            del self.ad.carrots



#
"""config - Configuration loading support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from configparser import ConfigParser
from os import environ, path

from .human_time import parse_timedelta
from .iso_8601 import parse_datetime, parse_delta
from .xdg_basedir import get_configs


def read_configs(__pkg: str, __name: str = 'config', *,
                 local: bool = True) -> ConfigParser:
    """Process configuration file stack.

    We export the time parsing functionality of ``jnrbase`` as custom
    converters for :class:`configparser.ConfigParser`:

    ===================  ===========================================
    Method               Function
    ===================  ===========================================
    ``.getdatetime()``   :func:`~jnrbase.iso_8601.parse_datetime`
    ``.gethumantime()``  :func:`~jnrbase.human_time.parse_timedelta`
    ``.gettimedelta()``  :func:`~jnrbase.iso_8601.parse_delta`
    ===================  ===========================================

    Args:
        __pkg: Package name to use as base for config files
        __name: File name to search for within config directories
        local: Whether to include config files from current directory
    Returns:
        Parsed configuration files
    """
    configs = get_configs(__pkg, __name)
    if local:
        localrc = path.abspath('.{}rc'.format(__pkg))
        if path.exists(localrc):
            configs.append(localrc)

    cfg = ConfigParser(converters={
        'datetime': parse_datetime,
        'humandelta': parse_timedelta,
        'timedelta': parse_delta,
    })
    cfg.read(configs, 'utf-8')
    cfg.configs = configs

    if 'NO_COLOUR' in environ or 'NO_COLOR' in environ:
        cfg.colour = False
    elif __pkg in cfg:
        if 'colour' in cfg[__pkg]:
            cfg.colour = cfg[__pkg].getboolean('colour')
        if 'color' in cfg[__pkg]:
            cfg.colour = cfg[__pkg].getboolean('color')
    else:
        cfg.colour = True

    return cfg



#
"""test_config - Test configuration support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from io import StringIO

from pytest import mark

from jnrbase import config
from jnrbase.context import chdir

from .utils import func_attr


@mark.parametrize('local,count', [
    (True, 4),
    (False, 3),
])
def test_config_loading(local, count, monkeypatch, path_exists_force):
    monkeypatch.setattr('builtins.open', lambda s, encoding: StringIO(''))
    monkeypatch.setenv('XDG_CONFIG_DIRS', 'test1:test2')
    cfg = config.read_configs('jnrbase', local=local)
    assert len(cfg.configs) == count
    if local:
        assert '/.jnrbaserc' in cfg.configs[-1]
    else:
        assert '/.jnrbaserc' not in cfg.configs


@func_attr('exists_result', False)
def test_config_loading_missing_files(path_exists_force):
    assert config.read_configs('jnrbase').configs == []


@mark.parametrize('envvar', ['NO_COLOUR', 'NO_COLOR'])
def test_no_colour_from_env(envvar, monkeypatch):
    monkeypatch.setenv(envvar, 'set')
    cfg = config.read_configs('jnrbase')
    assert not cfg.colour


def test_colour_default(monkeypatch):
    monkeypatch.setattr('os.environ', {})
    cfg = config.read_configs('jnrbase')
    assert cfg.colour


@mark.parametrize('directory', ['config', 'config2'])
def test_colour_from_config(directory, monkeypatch):
    monkeypatch.setattr('os.environ', {})
    with chdir('tests/data/{}'.format(directory)):
        cfg = config.read_configs('jnrbase', local=True)
    assert not cfg.colour



#
"""json_datetime - JSON datetime support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

import datetime
import json

from contextlib import suppress
from functools import partial, singledispatch, wraps
from typing import Any, Dict

from .iso_8601 import (format_datetime, format_delta, parse_datetime,
                       parse_delta)


encoder = json.JSONEncoder()  # pylint: disable=invalid-name


@singledispatch
def json_serialise(__o: Any) -> str:
    """Custom JSON serialiser.

    This simply falls through to :meth:`~json.JSONEncoder.default` when there
    isn't a custom dispatcher register for a type.

    Args:
        __o: Object to encode
    Returns:
        JSON-encoded string
    """
    return encoder.default(__o)


@json_serialise.register(datetime.datetime)
def datetime_serialise(__o: datetime.datetime) -> str:
    """JSON serialiser for ``datetime`` objects"""
    return format_datetime(__o)


@json_serialise.register(datetime.timedelta)
def timedelta_serialise(__o: datetime.timedelta) -> str:
    """JSON serialiser for ``timedelta`` objects"""
    return format_delta(__o)


def json_using_iso8601(__obj: Dict) -> Dict:
    """Parse ISO-8601 values from JSON databases.

    See :class:`json.JSONDecoder`

    Args:
        __obj: Object to decode
    """
    for key, value in __obj.items():
        with suppress(TypeError, ValueError):
            __obj[key] = parse_datetime(value)
        with suppress(TypeError, ValueError):
            __obj[key] = parse_delta(value)
    return __obj


# pylint: disable=invalid-name
dump = wraps(json.dump)(partial(json.dump, indent=4, default=json_serialise))
dumps = wraps(json.dumps)(partial(json.dumps, indent=4,
                                  default=json_serialise))
load = wraps(json.load)(partial(json.load, object_hook=json_using_iso8601))
loads = wraps(json.loads)(partial(json.loads, object_hook=json_using_iso8601))



#
"""test_debug - Test debug support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from operator import add

from pytest import mark, raises

from jnrbase.debug import DebugPrint, noisy_wrap, on_enter, on_exit, sys


@mark.parametrize('ftype', [
    on_enter,
    on_exit,
])
def test_decorator_no_message(ftype, capsys):
    @ftype
    def func(x, y):
        return x + y
    assert func(4, 3) == 7
    assert "{}ing 'func'({!r})".format(ftype.__name__[3:].capitalize(),
                                       func.__closure__[0].cell_contents) \
        in capsys.readouterr()[0]


@mark.parametrize('ftype', [
    on_enter,
    on_exit,
])
def test_decorator_with_message(ftype, capsys):
    assert ftype('custom message')(add)(4, 3) == 7
    assert 'custom message' in capsys.readouterr()[0]


@mark.parametrize('ftype', [
    on_enter,
    on_exit,
])
def test_decorator_with_failure(ftype, capsys):
    @ftype('custom message')
    def func(x, y):
        raise ValueError('boom')
    with raises(ValueError):
        func(4, 3)
    assert capsys.readouterr()[0] == 'custom message\n'


def test_DebugPrint(capsys):  # NOQA: N802
    DebugPrint.enable()
    try:
        print('boom')
        out, _ = capsys.readouterr()
        assert 'test_debug.py:' in out
        assert '] boom\n' in out
    finally:
        DebugPrint.disable()


def test_DebugPrint_no_stack_frame(capsys, monkeypatch):  # NOQA: N802
    monkeypatch.setattr('jnrbase.debug.inspect.currentframe', lambda: None)
    DebugPrint.enable()
    try:
        print('boom')
        assert 'unknown:000] boom\n' in capsys.readouterr()[0]
    finally:
        DebugPrint.disable()


def test_DebugPrint_double_toggle():  # NOQA: N802
    DebugPrint.enable()
    sys.stdout.first = True
    try:
        DebugPrint.enable()
        assert sys.stdout.first
    finally:
        DebugPrint.disable()
    assert not hasattr(sys.stdout, 'first')
    DebugPrint.disable()


def test_DebugPrint_decorator(capsys):  # NOQA: N802
    @noisy_wrap
    def func(x):
        print(hex(x))
        print(x)
    func(20)
    out, _ = capsys.readouterr()
    assert 'test_debug.py:' in out
    assert '] 0x14\n' in out
    assert '] 20\n' in out



#
"""pager - pager pipe support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from subprocess import run
from typing import Optional


# pylint: disable=redefined-outer-name
def pager(__text: str, *, pager: Optional[str] = 'less'):
    """Pass output through pager.

    See :manpage:`less(1)`, if you wish to configure the default pager.  For
    example, you may wish to check ``FRSX`` options.

    Args:
        __text: Text to page
        pager: Pager to use
    """
    if pager:
        run([pager, ], input=__text.encode())
    else:
        print(__text)



#
"""template - Jinja templating support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

import re
import sys

from datetime import datetime
from os import path
from typing import Callable, Union

import html2text as html2
import jinja2

from click import style
from pygments import highlight as pyg_highlight
from pygments.formatters import get_formatter_by_name
from pygments.lexers import get_lexer_by_name

from . import xdg_basedir
from .human_time import human_timestamp


#: Collection of custom filters to add to Jinja environment
FILTERS = {}


def jinja_filter(__func: Callable) -> Callable:
    """Simple decorator to add a new filter to Jinja environment.

    See also: :obj:`FILTERS`

    Args:
        __func: Function to add to Jinja environment
    Returns:
        Unmodified function
    """
    FILTERS[__func.__name__] = __func

    return __func


@jinja_filter
def colourise(__text: str, *args, **kwargs) -> str:
    """Colourise text using click’s style function.

    Returns text untouched if colour output is not enabled, or ``stdout`` is
    not a tty.

    See :func:`click.style` for parameters

    Args:
        __text: Text to colourise
    Returns:
        Colourised text, when possible
    """
    if sys.stdout.isatty():
        __text = style(__text, *args, **kwargs)
    return __text


@jinja_filter
def highlight(__text: str, *, lexer: str = 'diff',
              formatter: str = 'terminal') -> str:
    """Highlight text highlighted using ``pygments``.

    Returns text untouched if colour output is not enabled.

    See also: :pypi:`Pygments`

    Args:
        __text: Text to highlight
        lexer: Jinja lexer to use
        formatter: Jinja formatter to use
    Returns:
        Syntax highlighted output, when possible
    """
    if sys.stdout.isatty():
        lexer = get_lexer_by_name(lexer)
        formatter = get_formatter_by_name(formatter)
        __text = pyg_highlight(__text, lexer, formatter)
    return __text


@jinja_filter
def html2text(__html: str, *, width: int = 80,
              ascii_replacements: bool = False) -> str:
    """HTML to plain text renderer.

    See also: :pypi:`html2text`

    Args:
        __html: Text to process
        width: Paragraph width
        ascii_replacements: Use pseudo-ASCII replacements for Unicode
    Returns:
        Rendered text
    """
    html2.BODY_WIDTH = width
    html2.UNICODE_SNOB = ascii_replacements
    return html2.html2text(__html).strip()


@jinja_filter
def regexp(__string: str, __pattern: str, __repl: Union[Callable, str], *,
           count: int = 0, flags: int = 0) -> str:
    """Jinja filter for regexp replacements.

    See :func:`re.sub` for documentation.

    Returns:
        Text with substitutions applied
    """
    return re.sub(__pattern, __repl, __string, count, flags)


@jinja_filter
def relative_time(__timestamp: datetime) -> str:
    """Format a relative time.

    See :func:`~jnrbase.human_time.human_timestamp`

    Args:
        __timestamp: Event to generate relative timestamp against
    Returns:
        Human readable date and time offset
    """
    return human_timestamp(__timestamp)


def setup(__pkg: str) -> jinja2.Environment:
    """Configure a new Jinja environment with our filters.

    Args:
        __pkg: Package name to use as base for templates searches
    Returns:
        Configured Jinja environment
    """
    dirs = [path.join(d, 'templates')
            for d in xdg_basedir.get_data_dirs(__pkg)]

    env = jinja2.Environment(
        autoescape=jinja2.select_autoescape(['html', 'xml']),
        loader=jinja2.ChoiceLoader([jinja2.FileSystemLoader(s) for s in dirs]))
    env.loader.loaders.append(jinja2.PackageLoader(__pkg, 'templates'))
    env.filters.update(FILTERS)

    return env



#
"""colourise - Output colourisation support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

# The colouring we use here is hopefully quite self-explanatory, and is in
# common use in, for example, git

from click import echo, style


#: Global flag to disable *all* colourisation
COLOUR = True


def _colourise(text: str, colour: str) -> str:
    """Colour text, if possible.

    Args:
        text: Text to colourise
        colour: Colour to display text in
    Returns:
        Colourised text, if possible
    """
    if COLOUR:
        text = style(text, fg=colour, bold=True)
    return text


def fail(__text: str) -> str:
    """Format a failure message.

    Args:
        __text: Text to format
    Returns:
        Bright red text, if possible
    """
    return _colourise(__text, 'red')


def info(__text: str) -> str:
    """Format an informational message.

    Args:
        __text: Text to format
    Returns:
        Bright blue text, if possible
    """
    return _colourise(__text, 'blue')


def success(__text: str) -> str:
    """Format a success message.

    Args:
        __text: Text to format
    Returns:
        Bright green text, if possible
    """
    return _colourise(__text, 'green')


def warn(__text: str) -> str:
    """Format a warning message.

    Args:
        __text: Text to format
    Returns:
        Bright yellow text, if possible
    """
    return _colourise(__text, 'yellow')


def pfail(__text: str) -> None:  # pragma: no cover
    """Pretty print a failure message to ``stdout``.

    Args:
        __text: Text to print
    """
    echo(fail(__text))


def pinfo(__text: str) -> None:  # pragma: no cover
    """Pretty print an informational message to ``stdout``.

    Args:
        __text: Text to print
    """
    echo(info(__text))


def psuccess(__text: str) -> None:  # pragma: no cover
    """Pretty print a success message to ``stdout``.

    Args:
        __text: Text to print
    """
    echo(success(__text))


def pwarn(__text: str) -> None:  # pragma: no cover
    """Pretty print a warning message to ``stdout``.

    Args:
        __text: Text to print
    """
    echo(warn(__text))



#
"""attrdict - Dictionary with attribute access."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from typing import Any, Hashable


class AttrDict(dict):

    """Dictionary with attribute access.

    See also: :obj:`dict`
    """

    def __contains__(self, __key: Hashable) -> bool:
        """Check for item membership.

        Args:
            __key: Key to test for
        Returns:
            ``True``, if item in AttrDict
        """
        return hasattr(self, __key)

    def __getattr__(self, __key: Hashable) -> Any:
        """Support item access via dot notation.

        Args:
            __key: Key to fetch
        """
        try:
            return self[__key]
        except KeyError:
            raise AttributeError(__key)

    def __setattr__(self, __key: Hashable, __value: Any) -> None:
        """Support item assignment via dot notation.

        Args:
            __key: Key to set value for
            __value: Value to set key to
        """
        try:
            self[__key] = __value
        except Exception as err:
            raise AttributeError(str(err))

    def __delattr__(self, __key: Hashable) -> None:
        """Support item deletion via dot notation.

        Args:
            __key: Key to delete
        """
        try:
            del self[__key]
        except TypeError as err:
            raise AttributeError(str(err))


class ROAttrDict(AttrDict):

    """Read-only dictionary with attribute access.

    See also: :obj:`AttrDict`
    """

    def __setitem__(self, *args):
        """Handle attempt to modify read-only dictionary.

        Raises:
            AttributeError: On modification attempt
        """
        raise AttributeError('{!r} is read-only'.format(self.__class__))
    __delattr__ = __delitem__ = __setattr__ = __setitem__



#
"""test_context - Test path context handlers support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from os import getcwd, getenv
from re import escape
from subprocess import PIPE, run

from pytest import raises

from jnrbase import context


def test_chdir(tmpdir):
    orig = getcwd()
    dir_ = tmpdir.join('new').mkdir().strpath
    with context.chdir(dir_):
        assert getcwd() == dir_
    assert getcwd() == orig


def test_chdir_missing(tmpdir):
    with raises(FileNotFoundError, match=escape('[Errno 2]')), \
            context.chdir(tmpdir.join('missing_dir').strpath):
        pass


def test_env_override():
    assert getenv('SHELL') != 'hello'
    with context.env(SHELL='hello'):
        assert getenv('SHELL') == 'hello'
    assert getenv('SHELL') != 'hello'


def test_env_unset():
    assert getenv('SHELL')
    with context.env(SHELL=None):
        assert getenv('SHELL') is None
    assert getenv('SHELL')


def test_env_subshell_support():
    assert getenv('NOT_SET') is None
    with context.env(NOT_SET='hello'):
        # Switch to encoding kwarg when Python 3.5 support is dropped
        out = run(['printenv', ], stdout=PIPE).stdout.decode()
        assert 'NOT_SET=hello' in out.splitlines()
    assert getenv('NOT_SET') is None



#
"""httplib2_certs - httplib2 system certs finder."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

import sys
import warnings

from os import getenv, path

import httplib2


#: Allow fallback to bundled httplib2 certs.
#:
#: *Packagers*: Patch this to ``False``
ALLOW_FALLBACK = True

#: Default certificate locations for platforms
PLATFORM_FILES = {
    'linux': ['/etc/ssl/certs/ca-certificates.crt',
              '/etc/pki/tls/certs/ca-bundle.crt'],
    'freebsd': ['/usr/local/share/certs/ca-root-nss.crt', ],
}


def find_certs() -> str:
    """Find suitable certificates for ``httplib2``.

    Warning:
        The default behaviour is to fall back to the bundled certificates when
        no system certificates can be found.  If you're packaging ``jnrbase``
        *please* set ``ALLOW_FALLBACK`` to ``False`` to disable this very much
        unwanted behaviour, but please maintain the option so that downstream
        users can inspect the configuration easily.

    See also: :pypi:`httplib2`

    Returns:
        Path to SSL certificates
    Raises:
        RuntimeError: When no suitable certificates are found
    """
    bundle = path.realpath(path.dirname(httplib2.CA_CERTS))
    # Some distros symlink the bundled path location to the system certs
    if not bundle.startswith(path.dirname(httplib2.__file__)):
        return bundle
    for platform, files in PLATFORM_FILES.items():
        if sys.platform.startswith(platform):
            for cert_file in files:
                if path.exists(cert_file):
                    return cert_file
    # An apparently common environment setting for macOS users to workaround
    # the lack of “standard” certs installation
    if path.exists(getenv('CURL_CA_BUNDLE', '')):
        return getenv('CURL_CA_BUNDLE')
    if ALLOW_FALLBACK:
        warnings.warn('No system certs detected, falling back to bundled',
                      RuntimeWarning)
        return httplib2.CA_CERTS
    else:
        raise RuntimeError('No system certs detected!')



#
"""test_json_datetime - Test JSON datetime functions"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from datetime import datetime, timedelta, timezone

from pytest import raises

from jnrbase import json_datetime


def test_json_no_datetime():
    class Test:
        pass
    data = {'test': Test()}
    with raises(TypeError, match='not JSON serializable'):
        json_datetime.dumps(data, indent=None)


def test_json_datetime():
    data = {'test': datetime(2014, 2, 3, 18, 12, tzinfo=timezone.utc)}
    assert json_datetime.dumps(data, indent=None) == \
        '{"test": "2014-02-03T18:12:00Z"}'


def test_deep_json_datetime():
    data = {'test': [{'test2': datetime(2014, 2, 3, 18, 12,
                                        tzinfo=timezone.utc)}, ]}
    assert json_datetime.dumps(data, indent=None) == \
        '{"test": [{"test2": "2014-02-03T18:12:00Z"}]}'


def test_json_load_no_custom_dump():
    data = '{"test": "not a datetime", "not a string": 3}'
    assert json_datetime.loads(data) == {'test': 'not a datetime',
                                         'not a string': 3}


def test_json_timedelta():
    data = {'test': timedelta(hours=3, seconds=4)}
    assert json_datetime.dumps(data, indent=None) == '{"test": "PT03H04S"}'


def test_deep_json_timedelta():
    data = {'test': [{'test2': timedelta(hours=3, seconds=4)}, ]}
    assert json_datetime.dumps(data, indent=None) == \
        '{"test": [{"test2": "PT03H04S"}]}'


def test_roundtrip():
    data = {
        'datetime': datetime(2014, 2, 3, 18, 12, tzinfo=timezone.utc),
        'timedelta': timedelta(hours=3, seconds=4),
    }
    json = json_datetime.dumps(data, indent=None)
    assert json_datetime.loads(json) == data



#
"""cmdline - Command line functionality for jnrbase."""
# Copyright © 2017-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from configparser import NoOptionError, NoSectionError
from contextlib import suppress
from datetime import datetime, timezone
from inspect import signature
from io import TextIOBase
from subprocess import CalledProcessError, run
from typing import Callable, Optional

from click import (Context, File, argument, echo, group, option, pass_context,
                   version_option)

import jnrbase
from jnrbase import (_version, colourise, config, git, httplib2_certs,
                     human_time, i18n, iso_8601, json_datetime, pip_support,
                     template, timer, xdg_basedir)


_, N_ = i18n.setup(jnrbase)


def get_default(__func: Callable, __arg: str) -> str:
    """Fetch default value for a function argument

    Args:
        __func: Function to inspect
        __arg: Argument to extract default value for
    """
    return signature(__func).parameters[__arg].default


# pylint: disable=missing-docstring


@group(help=_('Possibly useful cli functionality.'),
       epilog=_('Please report bugs at '
                'https://github.com/JNRowe/jnrbase/issues'),
       context_settings={'help_option_names': ['-h', '--help']})
@version_option(_version.dotted)
def cli():
    pass


@cli.group(help=_('Format messages for users.'))
def messages():
    pass


def text_arg(__f: Callable) -> Callable:
    return argument('text')(__f)


@messages.command(help=_(colourise.fail.__doc__.splitlines()[0]))
@text_arg
@pass_context
def fail(ctx: Context, text: str):
    colourise.pfail(text)
    ctx.exit(1)


@messages.command(help=_(colourise.info.__doc__.splitlines()[0]))
@text_arg
def info(text: str):
    colourise.pinfo(text)


@messages.command(help=_(colourise.success.__doc__.splitlines()[0]))
@text_arg
def success(text: str):
    colourise.psuccess(text)


@messages.command(help=_(colourise.warn.__doc__.splitlines()[0]))
@text_arg
def warn(text: str):
    colourise.pwarn(text)


@cli.command(name='config', help=_('Extract or list values from config.'))
@option('-n', '--name', default=get_default(config.read_configs, '__name'),
        help=_('Config file to read from.'))
@option('-l', '--local / --no-local', help=_('Read local .<package>rc files.'))
@argument('package')
@argument('section')
@argument('key', required=False)
def config_(name: str, local: bool, package: str, section: str,
            key: Optional[str]):
    cfg = config.read_configs(package, name, local=local)
    if key:
        with suppress(NoOptionError, NoSectionError):
            echo(cfg.get(section, key))
    else:
        with suppress(NoSectionError):
            for opt in cfg.options(section):
                colourise.pinfo(opt)
                echo('    {}'.format(cfg.get(section, opt)))


@cli.command('find-tag', help=_('Find tag for git repository.'))
@option('-m', '--match', default=get_default(git.find_tag, '__matcher'),
        help=_('Limit the selection of matches with glob.'))
@option('-s', '--strict / --no-strict', help=_('Always generate a result.'))
@option('-d', '--directory', default=get_default(git.find_tag, 'git_dir'),
        help=_('Git repository to operate on.'))
def find_tag(match: str, strict: bool, directory: str):
    with suppress(CalledProcessError):
        echo(git.find_tag(match, strict=strict, git_dir=directory))


@cli.command(help=_('Find location of system certificates.'))
def certs():
    echo(httplib2_certs.find_certs())


@cli.command('pretty-time', help=_('Format timestamp for human consumption.'))
@argument('timestamp')
def pretty_time(timestamp: str):
    try:
        parsed = iso_8601.parse_datetime(timestamp)
    except ValueError:
        now = datetime.utcnow().replace(tzinfo=timezone.utc)
        try:
            delta = iso_8601.parse_delta(timestamp)
        except ValueError:
            delta = human_time.parse_timedelta(timestamp)
        parsed = now - delta

    echo(human_time.human_timestamp(parsed))


@cli.command('pip-requires', help=_('Parse pip requirements file.'))
@argument('name')
def pip_requires(name: str):
    requires = pip_support.parse_requires(name)
    for req in requires:
        echo(req)


@cli.command('gen-text', help=_('Create output from Jinja template.'))
@option('-e', '--env', type=File(),
        help=_('JSON data to generate output with.'))
@argument('package')
@argument('tmpl')
def gen_text(env: TextIOBase, package: str, tmpl: str):
    if env:
        env_args = json_datetime.load(env)
    else:
        env_args = {}
    jinja_env = template.setup(package)
    echo(jinja_env.get_template(tmpl).render(**env_args))


@cli.command(help=_('Time the output of a command'))
@argument('command')
@pass_context
def time(ctx: Context, command: str):
    with timer.Timing(verbose=True):
        proc = run(command, shell=True)
    ctx.exit(proc.returncode)


@cli.group(help=_('Query package directories'))
def dirs():
    pass


for k in ['cache', 'config', 'data']:
    @dirs.command(name=k,
                  help=_('Display {} dir honouring XDG basedir.'.format(k)))
    @argument('package')
    @pass_context
    def func(ctx: Context, package: str):
        echo(getattr(xdg_basedir, 'user_{}'.format(ctx.command.name))(package))


if __name__ == '__main__':
    cli()



#
"""pip_support - pip workarounds support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#                        Nathan McGregor <nathan.mcgregor@astrium.eads.net>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

# pip, both as a tool and a package manager, are not available on many of the
# systems I use.  However, lots of Python users like to use it so we’ll need to
# support the workflow to some extent…

import re

from os import path
from sys import version_info
from typing import List


def parse_requires(__fname: str) -> List[str]:
    """Parse ``pip``-style requirements files.

    This is a *very* naïve parser, but very few packages make use of the more
    advanced features.  Support for other features will be added only when
    packages in the wild depend on them.

    Args:
        __fname: Base file to pass
    Returns:
        Parsed dependencies
    """
    deps = []
    with open(__fname) as req_file:
        entries = [s.split('#')[0].strip() for s in req_file.readlines()]
        for dep in entries:
            if not dep:
                continue
            elif dep.startswith('-r '):
                include = dep.split()[1]
                if '/' not in include:
                    include = path.join(path.dirname(__fname), include)
                deps.extend(parse_requires(include))
                continue
            elif ';' in dep:
                dep, marker = [s.strip() for s in dep.split(';')]
                # Support for other markers will be added when they’re actually
                # found in the wild
                match = re.fullmatch(r"""
                        (?:python_version)  # Supported markers
                        \s*
                        (?:<=?|==|>=?)  # Supported comparisons
                        \s*
                        (?P<quote>(?:'|"))(?:[\d\.]+)(?P=quote)  # Test
                    """, marker, re.VERBOSE)
                if not match:
                    raise ValueError('Invalid marker {!r}'.format(marker))
                env = {
                    '__builtins__': {},
                    'python_version': '{}.{}'.format(*version_info[:2]),
                }
                if not eval(marker, env):  # pylint: disable=eval-used
                    continue
            deps.append(dep)
    return deps



#
"""test_cmdline - Test command line functionality support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from shutil import which

from click.testing import CliRunner
from pytest import mark

from jnrbase.cmdline import cli


pytestmark = mark.skipif(not which('git'), reason='Requires git')


@mark.parametrize('type_', ['fail', 'info', 'success', 'warn'])
def test_messages(type_):
    runner = CliRunner()
    result = runner.invoke(cli, ['messages', type_, 'test str'])
    if type_ == 'fail':
        assert result.exit_code == 1
    else:
        assert result.exit_code == 0
    assert 'test str' in result.output



#
"""git - Utilities for git support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from subprocess import CalledProcessError, check_output

from .context import chdir


def find_tag(__matcher: str = 'v[0-9]*', *, strict: bool = True,
             git_dir: str = '.') -> str:
    """Find closest tag for a git repository.

    Note:
        This defaults to `Semantic Version`_ tag matching.

    Args:
        __matcher: Glob-style tag pattern to match
        strict: Allow commit-ish, if no tag found
        git_dir: Repository to search
    Returns:
        Matching tag name

    .. _Semantic Version: http://semver.org/
    """
    command = 'git describe --abbrev=12 --dirty'.split()
    with chdir(git_dir):
        try:
            stdout = check_output(command + ['--match={}'.format(__matcher), ])
        except CalledProcessError:
            if strict:
                raise
            stdout = check_output(command + ['--always', ])

        stdout = stdout.decode('ascii', 'replace')
    return stdout.strip()



#
"""test_entry - Test entry point support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from pytest import raises

from jnrbase.entry import entry_point


def test_entrypoint():
    def f():
        return 42
    f.__module__ = '__main__'
    with raises(SystemExit, match='42'):
        entry_point(f)


def test_fallthrough():
    @entry_point
    def f():
        return 42
    assert f() == 42



#
"""test_template - Test Jinja template support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from datetime import datetime, timedelta

from pytest import mark

from jnrbase import template


def test_setup():
    env = template.setup('jnrbase')
    assert 'safe' in env.filters


def test_filter_decorator():
    @template.jinja_filter
    def test():
        return ''
    assert template.FILTERS['test'] == test


@mark.parametrize('filter_,args,kwargs,expected', [
    ('colourise', ('test', 'green'), {}, '\x1b[32mtest\x1b[0m'),
    ('regexp', ('test', 't', 'T'), {}, 'TesT'),
    ('highlight', ('f = lambda: True', ), {'lexer': 'python'},
     'f = \x1b[34mlambda\x1b[39;49;00m: \x1b[36mTrue\x1b[39;49;00m\n'),
    ('html2text', ('<b>test</b>', ), {}, '**test**'),
    ('relative_time', (datetime.utcnow() - timedelta(days=1), ), {},
     'yesterday'),
])
def test_custom_filter(filter_, args, kwargs, expected, monkeypatch):
    monkeypatch.setattr('sys.stdout.isatty', lambda: True)
    env = template.setup('jnrbase')
    assert env.filters[filter_](*args, **kwargs) == expected


@mark.parametrize('filter_,args,kwargs,expected', [
    ('colourise', ('test', 'green'), {}, 'test'),
    ('highlight', ('f = lambda: True', ), {'lexer': 'python'},
     'f = lambda: True'),
])
def test_custom_filter_fallthrough(filter_, args, kwargs, expected):
    env = template.setup('jnrbase')
    assert env.filters[filter_](*args, **kwargs) == expected



#
"""conf - Sphinx configuration information."""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

import os
import sys

from contextlib import suppress
from subprocess import PIPE, CalledProcessError, run

root_dir = os.path.dirname(os.path.dirname(__file__))
sys.path.insert(0, root_dir)

import jnrbase  # NOQA: E402

on_rtd = os.getenv('READTHEDOCS')
if not on_rtd:
    import sphinx_rtd_theme

# General configuration {{{
extensions = \
    ['sphinx.ext.{}'.format(ext)
     for ext in ['autodoc', 'coverage', 'doctest', 'extlinks', 'intersphinx',
                 'napoleon', 'todo', 'viewcode']] \
    + ['sphinxcontrib.{}'.format(ext) for ext in []] \
    + ['sphinx_autodoc_typehints', ]

if not on_rtd:
    # Only activate spelling if it is installed.  It is not required in the
    # general case and we don’t have the granularity to describe this in a
    # clean way
    try:
        from sphinxcontrib import spelling  # NOQA: F401
    except ImportError:
        pass
    else:
        extensions.append('sphinxcontrib.spelling')

master_doc = 'index'

rst_epilog = """
.. |PyPI| replace:: :abbr:`PyPI (Python Package Index)`
.. |modref| replace:: :mod:`jnrbase`
"""

default_role = 'any'

needs_sphinx = '1.6'
needs_extensions = {
    'sphinx-autodoc-typehints': '1.2'
}

nitpicky = True
# }}}

# Project information {{{
project = 'jnrbase'
copyright = '2014-2018  James Rowe'

version = '{major}.{minor}'.format_map(jnrbase._version.dict)
release = jnrbase._version.dotted

modindex_common_prefix = ['jnrbase.', ]

trim_footnote_reference_space = True
# }}}

# Options for HTML output {{{
# readthedocs.org handles this setup for their builds, but it is nice to see
# approximately correct builds on the local system too
if not on_rtd:
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path(), ]

with suppress(CalledProcessError):
    proc = run(['git', 'log', "--pretty=format:'%ad [%h]'", '--date=short',
                '-n1'],
               stdout=PIPE)
    html_last_updated_fmt = proc.stdout.decode()

html_copy_source = False

html_experimental_html5_writer = True
# }}}

# Options for manual page output {{{
man_pages = [
    ('jnrbase.1', 'jnrbase', 'jnrbase Documentation', ['James Rowe', ], 1)
]
# }}}

# autodoc extension settings {{{
autoclass_content = 'both'
autodoc_default_flags = ['members', ]
# }}}

# coverage extension settings {{{
coverage_write_headline = False
# }}}

# extlinks extension settings {{{
github_base = 'https://github.com/JNRowe/{}/'.format(project)
extlinks = {
    'issue': ('{}issues/%s'.format(github_base), 'issue #'),
    'pr': ('{}pull/%s'.format(github_base), 'pull request #'),
    'pypi': ('https://pypi.python.org/pypi/%s', ''),
}
# }}}

# graphviz extension settings {{{
graphviz_output_format = 'svg'
# }}}

# intersphinx extension settings {{{
intersphinx_mapping = {
    k: (v, os.getenv('SPHINX_{}_OBJECTS'.format(k.upper())))
    for k, v in {
        'click': 'http://click.pocoo.org/6/',
        'jinja2': 'http://jinja.pocoo.org/docs/',
        'pygments': 'http://pygments.org/',
        'python': 'http://docs.python.org/3/',
    }.items()
}
# }}}

# napoleon extension settings {{{
napoleon_numpy_docstring = False
# }}}

# spelling extension settings {{{
spelling_lang = 'en_GB'
spelling_word_list_filename = 'wordlist.txt'
# }}}

# todo extension settings {{{
todo_include_todos = True
# }}}



#
"""test_iso_8601 - Test ISO-8601 handling functions"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from datetime import datetime, timedelta, timezone

from hiro import Timeline
from pytest import mark, raises

from jnrbase.iso_8601 import (format_datetime, format_delta, parse_datetime,
                              parse_delta)


@mark.parametrize('string,expected', [
    ('2011-05-04T08:00:00Z', datetime(2011, 5, 4, 8, 0, tzinfo=timezone.utc)),
    ('2011-05-04T09:15:00Z', datetime(2011, 5, 4, 9, 15, tzinfo=timezone.utc)),
    ('2011-05-04T09:15:00', datetime(2011, 5, 4, 9, 15, tzinfo=timezone.utc)),
    ('', None),
])
def test_parse_datetime(string, expected):
    if expected is None:
        with Timeline().freeze():
            now = datetime.now(timezone.utc)
            assert parse_datetime(string) == now
    else:
        assert parse_datetime(string) == expected


@mark.parametrize('dt,expected', [
    (datetime(2011, 5, 4, 8, 0, tzinfo=timezone.utc), '2011-05-04T08:00:00Z'),
    (datetime(2011, 5, 4, 9, 15, tzinfo=timezone.utc), '2011-05-04T09:15:00Z'),
])
def test_format_datetime(dt, expected):
    assert format_datetime(dt) == expected


@mark.parametrize('string,expected', [
    ('PT04H30M21S', timedelta(hours=4, minutes=30, seconds=21)),
    ('PT00H12M01S', timedelta(minutes=12, seconds=1)),
    ('PT00H12M01.45S', timedelta(minutes=12, seconds=1, microseconds=45)),
    ('PT04H', timedelta(hours=4)),
    ('PT04H30M', timedelta(hours=4, minutes=30)),
    ('PT30M', timedelta(minutes=30)),
    ('PT04H21S', timedelta(hours=4, seconds=21)),
    ('PT4H', timedelta(hours=4)),
    ('P3DT04H', timedelta(days=3, hours=4)),
    ('P3D', timedelta(days=3)),
])
def test_parse_duration(string, expected):
    assert parse_delta(string) == expected


@mark.parametrize('delta,expected', [
    (timedelta(hours=4, minutes=30, seconds=21), 'PT04H30M21S'),
    (timedelta(minutes=12, seconds=1), 'PT12M01S'),
    (timedelta(), ''),
    (timedelta(days=3, hours=4), 'P3DT04H'),
    (timedelta(days=3), 'P3D'),
    (timedelta(days=2, hours=22), 'P2DT22H'),
    (timedelta(hours=4), 'PT04H'),
    (timedelta(hours=4, minutes=30), 'PT04H30M'),
    (timedelta(minutes=30), 'PT30M'),
    (timedelta(hours=4, seconds=21), 'PT04H21S'),
])
def test_format_duration(delta, expected):
    assert format_delta(delta) == expected


def test_parse_null_duration():
    assert parse_delta('') == timedelta()


def test_parse_invalid_duration():
    with raises(ValueError, match='Unable to parse delta '):
        parse_delta('P3Dwhoops')



#
"""entry - Simple, lazy, module executing support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from typing import Callable, NoReturn


def entry_point(__func: Callable) -> Callable[..., NoReturn]:
    """Execute function when module is run directly.

    Note:
        This allows fall through for importing modules that use it.

    Args:
        __func: Function to run
    """
    if __func.__module__ == '__main__':
        import sys
        sys.exit(__func())
    else:
        return __func



#
"""human_time - Handle human readable date formats."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

import datetime
import re


def human_timestamp(__timestamp: datetime.datetime) -> str:
    """Format a relative time.

    Args:
        __timestamp: Event to generate relative timestamp against
    Returns:
        Human readable date and time offset
    """
    numstr = '. a two three four five six seven eight nine ten'.split()

    matches = [
        60 * 60 * 24 * 365,
        60 * 60 * 24 * 28,
        60 * 60 * 24 * 7,
        60 * 60 * 24,
        60 * 60,
        60,
        1,
    ]
    match_names = ['year', 'month', 'week', 'day', 'hour', 'minute', 'second']

    if __timestamp.tzinfo is None:
        __timestamp = __timestamp.replace(tzinfo=datetime.timezone.utc)

    now = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)

    delta = int((now - __timestamp).total_seconds())
    for scale in matches:
        i = delta // scale
        if i:
            name = match_names[matches.index(scale)]
            break
    else:
        i = 0  # Too small

    if i == 0:
        result = 'right now'
    elif i == 1 and name in ('year', 'month', 'week'):
        result = 'last {}'.format(name)
    elif i == 1 and name == 'day':
        result = 'yesterday'
    elif i == 1 and name == 'hour':
        result = 'about an hour ago'
    else:
        result = 'about {} {}{} ago'.format(i if i > 10 else numstr[i], name,
                                            's' if i > 1 else '')
    return result


def parse_timedelta(__delta: str) -> datetime.timedelta:
    """Parse human readable frequency.

    Args:
        __delta: Frequency to parse
    """
    match = re.fullmatch(r"""
            ^(\d+(?:|\.\d+))  # value, possibly float
            \ *
            ([hdwmy])$  # units
         """, __delta, re.IGNORECASE | re.VERBOSE)
    if not match:
        raise ValueError('Invalid ‘frequency’ value')
    value, units = match.groups()
    units_i = 'hdwmy'.index(units.lower())
    # hours per hour/day/week/month/year
    multiplier = (1, 24, 168, 672, 8760)
    return datetime.timedelta(hours=float(value) * multiplier[units_i])



#
"""test_httplib2_certs - Test cert locating functions"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from pytest import mark, raises, warns

from jnrbase import httplib2_certs

from .utils import func_attr


exists_result = lambda x: func_attr('exists_result', x)  # NOQA: E731


def test_upstream_import(path_exists_force):
    import ca_certs_locater
    assert ca_certs_locater.get() == '/etc/ssl/certs/ca-certificates.crt'


def test_unbundled_package_import(monkeypatch):
    monkeypatch.setattr('jnrbase.httplib2_certs.httplib2.CA_CERTS',
                        '/fixed_by_distributor/certs.crt')
    assert httplib2_certs.find_certs() == '/fixed_by_distributor'


@exists_result(False)
def test_bundled(path_exists_force):
    with warns(RuntimeWarning) as record:
        httplib2_certs.find_certs()
    assert 'falling back' in record[0].message.args[0]


@exists_result(False)
def test_bundled_fail(monkeypatch, path_exists_force):
    monkeypatch.setattr('jnrbase.httplib2_certs.ALLOW_FALLBACK', False)
    with raises(RuntimeError, match='No system certs detected!'):
        httplib2_certs.find_certs()


def test_freebsd_paths(monkeypatch, path_exists_force):
    monkeypatch.setattr('sys.platform', 'freebsd')
    assert httplib2_certs.find_certs() \
        == '/usr/local/share/certs/ca-root-nss.crt'


@exists_result(False)
def test_freebsd_no_installed_certs(monkeypatch, path_exists_force):
    monkeypatch.setattr('sys.platform', 'freebsd')
    monkeypatch.setattr('jnrbase.httplib2_certs.ALLOW_FALLBACK', False)
    with raises(RuntimeError, match='No system certs detected!'):
        httplib2_certs.find_certs()


@mark.parametrize('file', [
    '/etc/ssl/certs/ca-certificates.crt',
    '/etc/pki/tls/certs/ca-bundle.crt',
])
def test_distros(file, monkeypatch):
    monkeypatch.setattr('jnrbase.httplib2_certs.path.exists',
                        lambda s: s == file)
    assert httplib2_certs.find_certs() == file


def test_curl_bundle(monkeypatch):
    monkeypatch.setattr('os.path.exists', lambda s: s == 'silly_platform_user')
    monkeypatch.setenv('CURL_CA_BUNDLE', 'silly_platform_user')
    assert httplib2_certs.find_certs() == 'silly_platform_user'



#! /usr/bin/env python3
"""setup.py - Setuptools tasks and config for jnrbase."""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from configparser import ConfigParser
from glob import glob
from importlib.util import module_from_spec, spec_from_file_location
from os import path
from types import ModuleType
from typing import List

from setuptools import setup
from setuptools.command.test import test


class PytestTest(test):
    def finalize_options(self):
        test.finalize_options(self)
        self.test_args = ['tests/', ]
        self.test_suite = True

    def run_tests(self):
        from sys import exit
        from pytest import main
        exit(main(self.test_args))


def import_file(package: str, fname: str) -> ModuleType:
    """Import file directly.

    This is a hack to import files from packages without importing
    <package>/__init__.py, its purpose is to allow import without requiring
    all the dependencies at this point.

    Args:
        package: Package to import from
        fname: File to import
    Returns:
        Imported module
    """
    mod_name = fname.rstrip('.py')
    spec = spec_from_file_location(mod_name, '{}/{}'.format(package, fname))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def make_list(s: str) -> List[str]:
    return s.strip().splitlines()


conf = ConfigParser()
conf.read('setup.cfg')
metadata = dict(conf['metadata'])

pip_support = import_file(metadata['name'], 'pip_support.py')

extras_require = {}
for file in glob('extra/requirements-*.txt'):
    suffix = path.splitext(file)[0].split('-')[1]
    if suffix not in ['dev', 'doc', 'test']:
        extras_require[suffix] = pip_support.parse_requires(file)

tests_require = pip_support.parse_requires('extra/requirements-test.txt')

metadata = dict(conf['metadata'])
for k in ['classifiers', 'packages', 'py_modules']:
    if k in metadata:
        metadata[k] = make_list(metadata[k])

for k in ['entry_points', 'package_data']:
    if k in metadata:
        metadata[k] = eval(metadata[k], {'__builtins__': {}})

with open('README.rst') as readme:
    metadata['long_description'] = readme.read()

_version = import_file(metadata['name'], '_version.py')

setup(
    version=_version.dotted,
    extras_require=extras_require,
    tests_require=tests_require,
    cmdclass={'test': PytestTest},
    zip_safe=False,
    **metadata,
)



#
"""xdg_basedir - XDG base directory support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

# See http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html

import sys

from os import getenv, path
from typing import List


#: Allow macOS directory structure
ALLOW_DARWIN = True


def user_cache(__pkg: str) -> str:
    """Return a cache location honouring :envvar:`XDG_CACHE_HOME`.

    .. envvar:: XDG_CACHE_HOME

        See XDG base directory spec.

    Args:
        __pkg: Package name
    """
    if ALLOW_DARWIN and sys.platform == 'darwin':
        user_dir = '~/Library/Caches'
    else:
        user_dir = getenv('XDG_CACHE_HOME',
                          path.sep.join([getenv('HOME', ''), '.cache']))

    return path.expanduser(path.sep.join([user_dir, __pkg]))


def user_config(__pkg: str) -> str:
    """Return a config location honouring :envvar:`XDG_CONFIG_HOME`.

    .. envvar:: XDG_CONFIG_HOME

        See XDG base directory spec.

    Args:
        __pkg: Package name
    """
    if ALLOW_DARWIN and sys.platform == 'darwin':
        user_dir = '~/Library/Preferences'
    else:
        user_dir = getenv('XDG_CONFIG_HOME',
                          path.sep.join([getenv('HOME', ''), '.config']))
    return path.expanduser(path.sep.join([user_dir, __pkg]))


def user_data(__pkg: str) -> str:
    """Return a data location honouring :envvar:`XDG_DATA_HOME`.

    .. envvar:: XDG_DATA_HOME

        See XDG base directory spec.

    Args:
        __pkg: Package name
    """
    if ALLOW_DARWIN and sys.platform == 'darwin':
        user_dir = '~/Library/Application Support'
    else:
        user_dir = getenv('XDG_DATA_HOME',
                          path.sep.join([getenv('HOME', ''), '.local/share']))
    return path.expanduser(path.sep.join([user_dir, __pkg]))


def get_configs(__pkg: str, __name: str = 'config') -> List[str]:
    """Return all configs for given package.

    Args:
        __pkg: Package name
        __name: Configuration file name
    """
    dirs = [user_config(__pkg), ]
    dirs.extend(path.expanduser(path.sep.join([d, __pkg]))
                for d in getenv('XDG_CONFIG_DIRS', '/etc/xdg').split(':'))
    configs = []
    for dname in reversed(dirs):
        test_path = path.join(dname, __name)
        if path.exists(test_path):
            configs.append(test_path)
    return configs


def get_data(__pkg: str, __name: str) -> str:
    """Return top-most data file for given package.

    Args:
        __pkg: Package name
        __name: Data file name
    """
    dirs = [user_data(__pkg), ]
    dirs.extend(path.expanduser(path.sep.join([d, __pkg]))
                for d in getenv('XDG_DATA_DIRS',
                                '/usr/local/share/:/usr/share/').split(':'))
    for dname in dirs:
        test_path = path.join(dname, __name)
        if path.exists(test_path):
            return test_path
    raise FileNotFoundError('No data file {!r} for {!r}'.format(__name, __pkg))


def get_data_dirs(__pkg: str) -> List[str]:
    """Return all data directories for given package.

    Args:
        __pkg: Package name
    """
    dirs = [user_data(__pkg), ]
    dirs.extend(path.expanduser(path.sep.join([d, __pkg]))
                for d in getenv('XDG_DATA_DIRS',
                                '/usr/local/share/:/usr/share/').split(':'))
    return [d for d in dirs if path.isdir(d)]



#
"""iso_8601 - ISO-8601 support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

import datetime
import re

import ciso8601


def parse_delta(__string: str) -> datetime.timedelta:
    """Parse ISO-8601 duration string.

    Args:
        __string: Duration string to parse
    Returns:
        Parsed delta object
    """
    if not __string:
        return datetime.timedelta(0)
    match = re.fullmatch(r"""
        P
        ((?P<days>\d+)D)?
        T?
        ((?P<hours>\d{1,2})H)?
        ((?P<minutes>\d{1,2})M)?
        ((?P<seconds>\d{1,2})?((?:\.(?P<microseconds>\d+))?S)?)
    """, __string, re.VERBOSE)
    if not match:
        raise ValueError('Unable to parse delta {!r}'.format(__string))
    match_dict = {k: int(v) if v else 0 for k, v in match.groupdict().items()}
    return datetime.timedelta(**match_dict)


def format_delta(__timedelta: datetime.timedelta) -> str:
    """Format ISO-8601 duration string.

    Args:
        __timedelta: Duration to process
    Returns:
        ISO-8601 representation of duration
    """
    if __timedelta == datetime.timedelta(0):
        return ''
    days_s = '{}D'.format(__timedelta.days) if __timedelta.days else ''
    hours, minutes = divmod(__timedelta.seconds, 3600)
    minutes, seconds = divmod(minutes, 60)
    hours_s = '{:02d}H'.format(hours) if hours else ''
    minutes_s = '{:02d}M'.format(minutes) if minutes else ''
    seconds_s = '{:02d}S'.format(seconds) if seconds else ''
    return 'P{}{}{}{}{}'.format(days_s,
                                'T' if hours or minutes or seconds else '',
                                hours_s, minutes_s, seconds_s)


def parse_datetime(__string: str) -> datetime.datetime:
    """Parse ISO-8601 datetime string.

    Args:
        __string: Datetime string to parse
    Returns:
        Parsed datetime object
    """
    if not __string:
        datetime_ = datetime.datetime.now(datetime.timezone.utc)
    else:
        # pylint: disable=no-member
        datetime_ = ciso8601.parse_datetime(__string)
        if not datetime_:
            raise ValueError('Unable to parse timestamp {!r}'.format(__string))
    if datetime_.tzinfo is None:
        datetime_ = datetime_.replace(tzinfo=datetime.timezone.utc)
    return datetime_


def format_datetime(__datetime: datetime.datetime) -> str:
    """Format ISO-8601 datetime string.

    Args:
        __datetime: Datetime to process
    Returns:
        ISO-8601 compatible string
    """
    return __datetime.isoformat().replace('+00:00', 'Z')



#
"""debug - Miscellaneous debugging support."""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

import inspect
import os
import sys

from functools import wraps
from typing import Callable, Optional, TextIO, Union

_orig_stdout = sys.stdout  # pylint: disable=invalid-name


class DebugPrint:

    """Verbose print wrapper for debugging."""

    def __init__(self, __handle: TextIO) -> None:
        """Configure new DebugPrint handler.

        Args:
            __handle: File handle to override
        """
        self.handle = __handle

    def write(self, __text: str) -> None:
        """Write text to the debug stream.

        Args:
            __text: Text to write
        """
        if __text == os.linesep:
            self.handle.write(__text)
        else:
            frame = inspect.currentframe()
            if frame is None:
                filename = 'unknown'
                lineno = 0
            else:
                outer = frame.f_back
                filename = outer.f_code.co_filename.split(os.sep)[-1]
                lineno = outer.f_lineno
            self.handle.write('[{:>15s}:{:03d}] {}'.format(filename[-15:],
                                                           lineno, __text))

    @staticmethod
    def enable() -> None:
        """Patch ``sys.stdout`` to use ``DebugPrint``."""
        if not isinstance(sys.stdout, DebugPrint):
            sys.stdout = DebugPrint(sys.stdout)

    @staticmethod
    def disable() -> None:
        """Re-attach ``sys.stdout`` to its previous file handle."""
        sys.stdout = _orig_stdout


def noisy_wrap(__func: Callable) -> Callable:
    """Decorator to enable DebugPrint for a given function.

    Args:
        __func: Function to wrap
    Returns:
        Wrapped function

    """
    # pylint: disable=missing-docstring
    def wrapper(*args, **kwargs):
        DebugPrint.enable()
        try:
            __func(*args, **kwargs)
        finally:
            DebugPrint.disable()
    return wrapper


def on_enter(__msg: Optional[Union[Callable, str]] = None) -> Callable:
    """Decorator to display a message when entering a function.

    Args:
        __msg: Message to display
    Returns:
        Wrapped function

    """
    # pylint: disable=missing-docstring
    def decorator(__func):
        @wraps(__func)
        def wrapper(*args, **kwargs):
            if __msg:
                print(__msg)
            else:
                print('Entering {!r}({!r})'.format(__func.__name__, __func))
            return __func(*args, **kwargs)
        return wrapper
    if callable(__msg):
        return on_enter()(__msg)
    return decorator


def on_exit(__msg: Optional[Union[Callable, str]] = None) -> Callable:
    """Decorator to display a message when exiting a function.

    Args:
        __msg: Message to display
    Returns:
        Wrapped function

    """
    # pylint: disable=missing-docstring
    def decorator(__func):
        @wraps(__func)
        def wrapper(*args, **kwargs):
            try:
                return __func(*args, **kwargs)
            finally:
                if __msg:
                    print(__msg)
                else:
                    print('Exiting {!r}({!r})'.format(__func.__name__, __func))
        return wrapper
    if callable(__msg):
        return on_exit()(__msg)
    return decorator



#
"""test_xdg_basedir - Test XDG basedir support."""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from pytest import raises

from jnrbase import xdg_basedir

from .utils import func_attr


exists_result = lambda x: func_attr('exists_result', x)  # NOQA: E731


def test_cache_no_args(monkeypatch):
    monkeypatch.setenv('XDG_CACHE_HOME', '~/.xdg/cache')
    assert '/.xdg/cache/jnrbase' in xdg_basedir.user_cache('jnrbase')


def test_cache_no_home(monkeypatch):
    monkeypatch.setattr('os.environ', {})
    assert xdg_basedir.user_cache('jnrbase') == '/.cache/jnrbase'


def test_cache_macos(monkeypatch):
    monkeypatch.setattr('sys.platform', 'darwin')
    assert '/Caches' in xdg_basedir.user_cache('jnrbase')


def test_config_no_args(monkeypatch):
    monkeypatch.setenv('XDG_CONFIG_HOME', '~/.xdg/config')
    assert '/.xdg/config/jnrbase' in xdg_basedir.user_config('jnrbase')


def test_config_no_home(monkeypatch):
    monkeypatch.setattr('os.environ', {})
    assert xdg_basedir.user_config('jnrbase') == '/.config/jnrbase'


def test_data_no_args(monkeypatch):
    monkeypatch.setenv('XDG_DATA_HOME', '~/.xdg/local')
    assert '/.xdg/local/jnrbase' in xdg_basedir.user_data('jnrbase')


def test_data_no_home(monkeypatch):
    monkeypatch.setattr('os.environ', {})
    assert xdg_basedir.user_data('jnrbase') == '/.local/share/jnrbase'


def test_macos_paths(monkeypatch):
    monkeypatch.setattr('sys.platform', 'darwin')
    assert '/Library/Application Support/jnrbase' \
        in xdg_basedir.user_data('jnrbase')


@exists_result(False)
def test_get_configs_all_missing(path_exists_force):
    assert xdg_basedir.get_configs('jnrbase') == []


def test_get_configs(path_exists_force):
    assert len(xdg_basedir.get_configs('jnrbase')) == 2


def test_get_configs_custom_dirs(monkeypatch, path_exists_force):
    monkeypatch.setenv('XDG_CONFIG_DIRS', 'test1:test2')
    assert len(xdg_basedir.get_configs('jnrbase')) == 3


def test_get_configs_macos(monkeypatch, path_exists_force):
    monkeypatch.setattr('sys.platform', 'darwin')
    assert '/Library/' in xdg_basedir.get_configs('jnrbase')[-1]


@exists_result([True, False])
def test_get_data(monkeypatch, path_exists_force):
    monkeypatch.setenv('XDG_DATA_HOME', '~/.xdg/local')
    monkeypatch.setenv('XDG_DATA_DIRS', '/usr/share:test2')
    assert xdg_basedir.get_data('jnrbase', 'photo.jpg') \
        == '/usr/share/jnrbase/photo.jpg'


@exists_result(False)
def test_get_data_no_files(path_exists_force):
    with raises(FileNotFoundError, match='No data file'):
        xdg_basedir.get_data('jnrbase', 'photo.jpg')



#
"""test_timer - Test timer support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from datetime import timedelta

from hiro import Timeline

from jnrbase.timer import Timing


def test_timing():
    with Timeline() as timeline, Timing() as t:
        timeline.forward(3600)
    assert t.elapsed >= timedelta(hours=1)


def test_verbose_timing(capsys):
    with Timeline() as timeline, Timing(verbose=True) as t:
        timeline.forward(3600)
    assert t.elapsed >= timedelta(hours=1)
    assert 'Started about an hour ago' in capsys.readouterr()[0]



#
"""test_i18n - Test i18n setup functions"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from jnrbase import i18n


# This needs a *real* test suite, but its usage necessitates bundling a heap of
# test data and patching practically everything to make it work.


def test_translation_config():
    _, N_ = i18n.setup(i18n)
    assert _.__name__ == 'gettext'
    assert N_.__name__ == 'ngettext'



#
"""test_git - Test git repository support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from contextlib import contextmanager
from os import path
from shutil import which
from subprocess import CalledProcessError
from tarfile import open as open_tar
from tempfile import TemporaryDirectory

from pytest import mark, raises

from jnrbase.git import find_tag


pytestmark = mark.skipif(not which('git'), reason='Requires git')


@contextmanager
def tarball_data(tar_name):
    """Extract a tarball for test usage

    This fixture extracts a tarball, and returns the path to the extracted
    files.

    :see: `tar_name`
    """
    data_dir = path.join(path.dirname(__file__), 'data', 'git')
    with open_tar(path.join(data_dir, tar_name + '.tar'), 'r:') as tar:
        with TemporaryDirectory() as temp_dir:
            tar.extractall(temp_dir)
            yield str(path.join(temp_dir, tar_name))


def test_empty_repo():
    with tarball_data('empty') as tree, \
            raises(CalledProcessError, match='status 128'):
        find_tag(git_dir=tree)


def test_semver_repo():
    with tarball_data('semver') as tree:
        assert find_tag(git_dir=tree) == 'v2.3.4'


def test_non_strict():
    with tarball_data('empty') as tree:
        assert find_tag(strict=None, git_dir=tree) == 'db3ed35e8734'


def test_custom_match():
    with tarball_data('funky_names') as tree:
        assert find_tag('prefix[0-9]*', git_dir=tree) == 'prefix9.8.7.6'



#
"""test_pip_support - Test pip workarounds support"""
# Copyright © 2014-2017  James Rowe <jnrowe@gmail.com>
#
# This file is part of jnrbase.
#
# jnrbase is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# jnrbase is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# jnrbase.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0+

from os import path

from pytest import mark, raises

from jnrbase.pip_support import parse_requires


DATA_DIR = path.join(path.dirname(__file__), 'data', 'pip')


def data_file(fname):
    return path.join(DATA_DIR, fname)


def test_empty_parse():
    assert parse_requires(data_file('empty.txt')) == []


def test_comment_skipping():
    assert parse_requires(data_file('comments.txt')) == ['httplib2', 'lxml']


def test_include():
    assert parse_requires(data_file('base.txt')) == ['httplib2', 'lxml']


def test_abs_include():
    assert parse_requires(data_file('base_abs.txt')) == ['httplib2', 'lxml']


@mark.parametrize('version,expected', [
    ((3, 3, 6), ['contextlib2>=0.5.4', ]),
    ((3, 5, 0), []),
])
def test_parse_markers(version, expected, monkeypatch):
    monkeypatch.setattr('jnrbase.pip_support.version_info', version)
    assert parse_requires(data_file('markers.txt')) == expected


def test_invalid_markers():
    with raises(ValueError, match='Invalid marker'):
        parse_requires(data_file('invalid_markers.txt'))
